#include "Common.h"
#include "Utils.h"
#include "Plants.h"
#include "PlantBase.h"
#include "resource.h"
#include <stdio.h>

using namespace oapi;

// ==============================================================
// Plants module implementation.
// ==============================================================

Plants::Plants(HINSTANCE hDLL) : Module (hDLL) {

	_selectedCBody = NULL;
	_focusObject = NULL;
	_selectedPlant = PlantBase::Null;
	_selectedDepot = DepotBase::Null;
};

Plants::~Plants() {

}

// ==============================================================
// Load product definitions from cfg file.

void Plants::LoadProducts() {

	FILEHANDLE file = oapiOpenFile("Products.cfg", FILE_IN, CONFIG);
	if (file != NULL) {

		char *line;
		char buf[100];
		string value;
		while (oapiReadScenario_nextline(file, line)) {
			if (!strnicmp(line, "BEGIN_PRODUCT", 13)) {
				_products.push_back(Product::Create());
			} else if (!strnicmp(line, "ID", 2)) {
				value = line;
				_products.back()->SetId(value.substr(3));
			} else if (!strnicmp(line, "NAME", 4)) {
				value = line;
				_products.back()->SetName(value.substr(5));
			} else if (!strnicmp(line, "TYPE", 4)) {
				value = line;
				_products.back()->SetType(value.substr(5));
			} else if (!strnicmp(line, "VESSEL", 6)) {
				value = line;
				_products.back()->SetVessel(value.substr(7));
			} else if (!strnicmp(line, "SLOT_SIZE", 9)) {
				double length, width;
				sscanf(line + 9, "%lf %lf", &length, &width);
				_products.back()->SetSlotLength(length);
				_products.back()->SetSlotWidth(width);
			} else if (!strnicmp(line, "RESOURCE", 8)) {
				int amount;
				sscanf(line + 8, "%s %d", &buf, &amount);
				_products.back()->AddResource(buf, amount);
			}
		}

		oapiCloseFile(file, FILE_IN);
	}
}

// ==============================================================
// Load plants from cfg file if it exists for the given cBody

void Plants::LoadCBodyFacilities(FacilityType facilityType, OBJHANDLE hCBody) {

	char cBodyName[50];
	oapiGetObjectName(hCBody, cBodyName, 50);

	char fileName[70];
	sprintf(fileName, "Config\\%s\\%s.cfg", cBodyName, facilityType == PLANT_BASE ? "Plants" : "Depots");

	if (_access(fileName, 0) != -1) {
		FILEHANDLE file = oapiOpenFile(fileName, FILE_IN);
		if (file != NULL) {

			vector<DepotBasePtr> cBodyLocations;

			char *line;
			string value;
			while (oapiReadScenario_nextline(file, line)) {
				if (!strnicmp(line, "BEGIN_PLANT", 11)) {
					cBodyLocations.push_back(PlantBase::Create(hCBody));
				} else if (!strnicmp(line, "BEGIN_DEPOT", 11)) {
					cBodyLocations.push_back(DepotBase::Create(hCBody));
				} else if (!strnicmp(line, "NAME", 4)) {
					value = line;
					cBodyLocations.back()->SetName(value.substr(5));
				} else if (!strnicmp(line, "BASE", 4)) {
					value = line;
					cBodyLocations.back()->SetBaseName(value.substr(5));
				} else if (!strnicmp(line, "PRODUCT", 7)) {
					if (cBodyLocations.back()->IsPlant()) {
						value = line;
						PlantBase::Cast(cBodyLocations.back())->SetProduct(GetProduct(value.substr(8)));
					}
				} else if (!strnicmp(line, "RATE", 4)) {
					if (cBodyLocations.back()->IsPlant()) {
						double rate;
						sscanf(line + 4, "%lf", &rate);
						PlantBase::Cast(cBodyLocations.back())->SetRate(rate);
					}
				} else if (!strnicmp(line, "STORAGE", 7)) {
					ExpositionBase expos;
					StoragePtr storage = Storage::Create();
					char productId[50], defaultVesselName[50];
					int capacity;
					sscanf(line + 7, "%s %s %d %lf %lf %lf", &productId, &defaultVesselName, &capacity, 
						&(expos.lng), &(expos.lat), &(expos.hdg));
					storage->SetProduct(GetProduct(productId));
					storage->SetDefaultVesselName(defaultVesselName);
					storage->SetCapacity(capacity);
					cBodyLocations.back()->AddStorage(storage, expos);
				}
			}

			if (cBodyLocations.size() > 0) {
				if (facilityType == PLANT_BASE) {
					_plants.insert(make_pair(hCBody, cBodyLocations));
				} else {
					_depots.insert(make_pair(hCBody, cBodyLocations));
				}
			}

			oapiCloseFile(file, FILE_IN);
		}
	}
}

void Plants::ConsumeScenarioLine(string line) {

	_scenarioLines.push_back(line);
}

void Plants::SaveState(FILEHANDLE scn) {

	SaveFacilitiesState(PLANT_BASE, scn);
	SaveFacilitiesState(DEPOT_BASE, scn);
}

void Plants::SaveFacilitiesState(FacilityType facilityType, FILEHANDLE scn) {

	map<OBJHANDLE, vector<DepotBasePtr>>& facilities = GetFacilities(facilityType);

	char cBodyName[100];
	char buf[256];
	for (map<OBJHANDLE, vector<DepotBasePtr>>::iterator iter = facilities.begin(); iter != facilities.end(); ++iter) {
		oapiGetObjectName(iter->first, cBodyName, 100);
		vector<DepotBasePtr>& cBodyFacilities = iter->second;
		for (vector<DepotBasePtr>::iterator pIter = cBodyFacilities.begin(); pIter != cBodyFacilities.end(); ++pIter) {
			DepotBasePtr depot = *pIter;
			int storageCount = depot->GetStorageCount();
			string stored;
			for (int i = 0; i < storageCount; i++) {
				if (stored.size() > 0) {
					stored.append(":");
				}
				sprintf(buf, "%d", depot->GetStorage(i)->GetStored());
				stored.append(buf);
			}
			if (facilityType == PLANT_BASE) {
				sprintf(buf, "%s:%s:%s:%lf", cBodyName, depot->GetName().data(), stored.data(), 
					PlantBase::Cast(depot)->GetProgress());
			} else {
				sprintf(buf, "%s:%s:%s", cBodyName, depot->GetName().data(), stored.data());
			}
			oapiWriteLine(scn, buf);
		}
	}
}

// ==============================================================
// Init on simulation start.

void Plants::clbkSimulationStart(RenderMode mode) {

	LoadProducts();

	// create plants
	DWORD cBodyCount = oapiGetGbodyCount();
	for (DWORD i = 0; i < cBodyCount; i++) {
		LoadCBodyFacilities(PLANT_BASE, oapiGetGbodyByIndex(i));
		LoadCBodyFacilities(DEPOT_BASE, oapiGetGbodyByIndex(i));
	}

	// parse state from scenario lines
	vector<string> tokens;
	string line;
	int stored;
	double progress;
	for (vector<string>::iterator iter = _scenarioLines.begin(); iter != _scenarioLines.end(); ++iter) {

		line = *iter;
		Tokenize(line, tokens, ":");

		if (tokens.size() > 2) {
			DepotBasePtr facility = GetFacility(PLANT_BASE, tokens[0], tokens[1]);
			if (!facility) {
				facility = GetFacility(DEPOT_BASE, tokens[0], tokens[1]);
			}
			if (facility) {
				int storageCount = facility->GetStorageCount();
				int allParamsCount = facility->IsPlant() ? storageCount + 3 : storageCount + 2;
				if (tokens.size() == allParamsCount) {
					for (int i = 0; i < storageCount; i++) {
						sscanf(tokens[2 + i].data(), "%d", &stored);
						facility->GetStorage(i)->SetStored(stored);
					}
					if (facility->IsPlant()) {
						sscanf(tokens[tokens.size() - 1].data(), "%lf", &progress);
						PlantBase::Cast(facility)->SetProgress(progress);
					}
				}
			}
		}
		tokens.clear();
	}

	_productionDt = 0.0;
}

// ==============================================================
// Clear all data. Must do it because ExitModule() isn't called after finishing a simulation.

void Plants::clbkSimulationEnd() {

	_selectedPlant = PlantBase::Null;
	_selectedDepot = DepotBase::Null;
	_selectedCBody = NULL;

	// delete plants
#ifndef USE_SMART_PTRS
	for (map<OBJHANDLE, vector<DepotBase*>>::iterator iter = _plants.begin(); iter != _plants.end(); ++iter) {
		vector<DepotBase*>& cBodyPlants = iter->second;
		while(!cBodyPlants.empty()) {
			PlantBase* plant = (PlantBase*)cBodyPlants.back();
			int storageCount = plant->GetStorageCount();
			for (int i = 0; i < storageCount; i++) {
				delete plant->GetStorage(i);
			}
			delete plant;
			cBodyPlants.pop_back();
		}
	}
#endif
	_plants.clear();

	// delete depots
#ifndef USE_SMART_PTRS
	for (map<OBJHANDLE, vector<DepotBase*>>::iterator iter = _depots.begin(); iter != _depots.end(); ++iter) {
		vector<DepotBase*>& cBodyDepots = iter->second;
		while(!cBodyDepots.empty()) {
			DepotBase* depot = cBodyDepots.back();
			int storageCount = depot->GetStorageCount();
			for (int i = 0; i < storageCount; i++) {
				delete depot->GetStorage(i);
			}
			delete depot;
			cBodyDepots.pop_back();
		}
	}
#endif
	_depots.clear();

#ifndef USE_SMART_PTRS
	while(!_products.empty()) {
		delete _products.back();
		_products.pop_back();
	}
#endif
	_products.clear();
	_scenarioLines.clear();
}

// ==============================================================
// Frame update.

void Plants::clbkPostStep(double simt, double simdt, double mjd) {

	_productionDt += simdt;
	if (_productionDt > PRODUCTION_STEP) {
		for (map<OBJHANDLE, vector<DepotBasePtr>>::iterator iter = _plants.begin(); iter != _plants.end(); ++iter) {
			vector<DepotBasePtr>& cBodyPlants = iter->second;
			for (vector<DepotBasePtr>::iterator pIter = cBodyPlants.begin(); pIter != cBodyPlants.end(); ++pIter) {
				PlantBase::Cast(*pIter)->Produce(_productionDt);
			}
		}
		_productionDt = 0.0;
		UpdateDialogPlantInfo();
	}

	if (_focusObject != NULL) {
		//oapiSetFocusObject(_focusObject);
		_focusObject = NULL;
	}
}

// ==============================================================
// Expose products at the selected plant.

int Plants::UnloadSelectedFacility(FacilityType facilityType, const int& storageIndex, const int& amount, 
		char* customName) {

	if (facilityType == PLANT_BASE) {
		if (_selectedPlant) {
			_focusObject = oapiGetFocusObject();
			return _selectedPlant->Unload(storageIndex, amount, customName);
		}
	} else {
		if (_selectedDepot) {
			_focusObject = oapiGetFocusObject();
			return _selectedDepot->Unload(storageIndex, amount, customName);
		}
	}
	return NO_SELECTED_FACILITY;
}

// ==============================================================
// Take product to the selected plant.

bool Plants::LoadSelectedFacility(FacilityType facilityType, const int& storageIndex) {

	if (facilityType == PLANT_BASE) {
		if (_selectedPlant) {
			return _selectedPlant->Load(storageIndex);
		}
	} else {
		if (_selectedDepot) {
			return _selectedDepot->Load(storageIndex);
		}
	}
	return false;
}

// ==============================================================
// Get a product by id.

ProductPtr Plants::GetProduct(string id) {

	for (vector<ProductPtr>::iterator iter = _products.begin(); iter != _products.end(); ++iter) {
		if (id.compare((*iter)->GetId()) == 0) {
			return *iter;
		}
	}
	return Product::Null;
}

map<OBJHANDLE, vector<DepotBasePtr>>& Plants::GetFacilities(FacilityType facilityType) {

	return facilityType == PLANT_BASE ? _plants : _depots;
}

// ==============================================================
// Get a facility by names of cBody and plant.

DepotBasePtr Plants::GetFacility(FacilityType facilityType, const string& cBodyName, const string& facilityName) {

	map<OBJHANDLE, vector<DepotBasePtr>>& facilities = GetFacilities(facilityType);

	char name[100];
	for (map<OBJHANDLE, vector<DepotBasePtr>>::iterator iter = facilities.begin(); iter != facilities.end(); ++iter) {
		oapiGetObjectName(iter->first, name, 100);
		if (cBodyName.compare(name) == 0) {
			vector<DepotBasePtr>& cBodyFacilites = iter->second;
			for (vector<DepotBasePtr>::iterator pIter = cBodyFacilites.begin(); pIter != cBodyFacilites.end(); ++pIter) {
				if (facilityName.compare((*pIter)->GetName()) == 0) {
					return *pIter;
				}
			}
		}
	}

	return DepotBase::Null;
}

// ==============================================================
// Sets the selected cBody by its index.

bool Plants::SetSelectedCBodyIndex(FacilityType facilityType, int cBodyIndex) {

	map<OBJHANDLE, vector<DepotBasePtr>>& facilities = GetFacilities(facilityType);

	if (cBodyIndex < facilities.size() && cBodyIndex >= 0) {
		map<OBJHANDLE, vector<DepotBasePtr>>::iterator iter = facilities.begin();
		std::advance(iter, cBodyIndex);
		
		_selectedCBody = iter->first;
		return true;
	} else {
		return false;
	}
}

// ==============================================================
// Returns plants of the selected cBody.

vector<DepotBasePtr>& Plants::GetSelectedCBodyFacilities(FacilityType facilityType) {

	return GetCBodyFacilities(facilityType, _selectedCBody);
}

// ==============================================================
// Returns all the plants of the cBody.

vector<DepotBasePtr>& Plants::GetCBodyFacilities(FacilityType facilityType, OBJHANDLE cBody) {

	return GetFacilities(facilityType)[cBody];
}

// ==============================================================
// Returns the number of cBodies which have at least one plant.

int Plants::GetCBodiesWithFacilitiesCount(FacilityType facilityType) {

	return GetFacilities(facilityType).size();
}

// ==============================================================
// Remembers the selected plant. Returns false if the index is out of bounds.

bool Plants::SetSelectedFacilityIndex(FacilityType facilityType, int index) {

	vector<DepotBasePtr>& cBodyFacilities = GetSelectedCBodyFacilities(facilityType);

	if (index >= 0 && index < cBodyFacilities.size()) {
		if (facilityType == PLANT_BASE) {
			_selectedPlant = PlantBase::Cast(cBodyFacilities.at(index));
		} else {
			_selectedDepot = cBodyFacilities.at(index);
		}
		return true;
	} else {
		if (facilityType == PLANT_BASE) {
			_selectedPlant = PlantBase::Null;
		} else {
			_selectedDepot = DepotBase::Null;
		}
		return false;
	}
}

PlantBasePtr Plants::GetSelectedPlant() {

	return _selectedPlant;
}

DepotBasePtr Plants::GetSelectedDepot() {

	return _selectedDepot;
}

DepotBasePtr Plants::GetSelectedFacility(FacilityType fType) {

	if (fType == PLANT_BASE) {
		return _selectedPlant;
	} else {
		return _selectedDepot;
	}
}

// ==============================================================
// Returns all cBodies which have at least one facility of specified type.

vector<OBJHANDLE> Plants::GetCBodies(FacilityType facilityType) {

	map<OBJHANDLE, vector<DepotBasePtr>>& facilities = GetFacilities(facilityType);
	vector<OBJHANDLE> cBodies;
	for (map<OBJHANDLE, vector<DepotBasePtr>>::iterator iter = facilities.begin(); iter != facilities.end(); ++iter) {
		cBodies.push_back(iter->first);
	}
	return cBodies;
}
